
//******************************************************************************
//  FILE..........: CASLINK.C
//  COPYRIGHT.....: Copyright (C) 1999 Alexey Podrezov
//  VERSION.......: 1.0b
//  DESCRIPTION...: Cassette Interface Emulation utility
//  NOTES.........: Encodes/decodes files to be transfered through MSX cassette interface
//  HISTORY.......:
//  PERSON          DATE         COMMENT
//  --------------- ------------ -----------------------------------------------
//  Podrezov        07/09/1999   Initial version created

//******************************************************************************


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>
#include "CASLINK.H"


// Variables -------------------------------------------------------------------

char InputFileName[13], OutputFileName[13];
char* FileNamePointer[12];
unsigned int WavSampleRate, SignalLevel, MSXInterfaceSpeed, LongHeaderLength, ShortHeaderLength, SampleScale, SpeedScale, k, l;
FILE *InputFileHandle, *OutputFileHandle;
unsigned char BitStream[11], *InputMemPointer, *OutputMemPointer, FirstPauseLength, MiddlePauseLength, FinalPauseLength, SilenceByte, EncodeDecode, FileType, AddressBuffer[6];
size_t InputFileLength, OutputFileLength, BytesRead, BytesWritten, FileOffset, ExtraBytes, InputFileLengthTmp;
struct {
	char WavHeaderID[4];	// 'RIFF'
   DWORD SampleLength;
   char MMFileType[4];	// 'WAVE'
  	char FormatID[4];	// 'fmt '
   DWORD FormatSize;
   WORD FormatTag;
   WORD NumChannels;
   DWORD SamplesPerSec;
   DWORD BytesPerSec;
	WORD Dummy;
   WORD BitsPerSample;
	char DataID[4];	// 'data'
   DWORD PureSampleLength;
   } WavHeader;

unsigned char BasicFileHeaderID[10] = {0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3};
unsigned char BinFileHeaderID[10] = {0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0};
unsigned char DataFileHeaderID[10] = {0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea};
unsigned char NameBuffer[CASFILENAMELENGTH], ch;
unsigned char BitEncodingLength, FileHeaderIDLength;
size_t i, a, b;
DWORD j, MoreExtraBytes;

// Patterns of 1200Hz and 2400Hz waves - samples for 11025Hz file. if bigger sampling rate is used, multiply by SampleScale variable
unsigned char ZeroBit[2][10] = {
	{LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE},
	{LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE}
   };
unsigned char SetBit[2][10] = {
	{LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE},
   {LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE}
   };

char ErrorList[][80] = {
   " ERROR: Incorrect parameter passed to command line...\n\n",								// 1
   " ERROR: Insufficient command line information. Need more options...\n\n",		// 2
   " ERROR: Incorrect value specified...\n\n",														// 3
   " ERROR: Failed to allocate memory for output buffer...\n\n", 								// 4
   " ERROR: Failed to allocate memory for input buffer...\n\n", 								// 5
   " ERROR: Failed to open input file...\n\n", 														// 6
   " ERROR: Failed to manipulate input file pointer...\n\n", 									// 7
   " ERROR: Input file can't be processed...\n\n", 												// 8
   " ERROR: Failed to read input file...\n\n", 														// 9
   " ERROR: Input file format is not supported...\n\n", 											// 10
   " CANCEL: Overwrite request canceled by user...\n\n", 										// 11
   " ERROR: Failed to create output file...\n\n", 		 											// 12
   " ERROR: Failed to write to output file...\n\n", 		 										// 13
	" STOP: This type of 32kb ROM is not supported...\n\n", 	  									// 14
	" STOP: This type of file is not supported in multifile mode...\n\n", 					// 15
	" STOP: Basic program should be detokenized for multifile mode...\n\n", 			 	// 16
	" STOP: Basic program can't be loaded after binary in multifile mode...\n\n",			// 17
   ""
   };

char FileFormat[][40] = {
	"BAS (tokenized basic program)",
   "BIN (binary program)",
   "ROM (image of ROM cartridge)",
   "BAS (detokenized basic program)",
   ""
   };

char ROMLoader[80] = {
   0xC3,0x09,0x90,0x00,0x00,0x00,0x00,0x00,0x00,0xF3,0x2A,0x07,0x90,0xE5,0x7C,0xE6,
   0xC0,0x11,0x00,0x00,0x21,0x0F,0xF0,0x28,0x06,0x11,0x00,0x40,0x21,0x0C,0xF3,0x3A,
   0xFF,0xFF,0x2F,0xA4,0x4F,0x0F,0x0F,0x0F,0x0F,0xA5,0xB1,0x32,0xFF,0xFF,0xDB,0xA8,
   0xA4,0x47,0x0F,0x0F,0x0F,0x0F,0xA5,0xB0,0xD3,0xA8,0xD5,0x2A,0x03,0x90,0xED,0x5B,
   0x05,0x90,0xEB,0xED,0x52,0x44,0x4D,0x21,0x50,0x90,0xD1,0xED,0xB0,0xFB,0xC9,0x00
   };

char ROM32KLoader1[128] = {
   0xC3,0x1E,0x90,0x00,0x00,0x00,0x00,0x00,0x00,0x42,0x4C,0x4F,0x41,0x44,0x22,0x43,
   0x41,0x53,0x3A,0x52,0x4F,0x4D,0x33,0x32,0x4B,0x22,0x2C,0x52,0x0D,0x0D,0xF3,0x3A,
   0xFF,0xFF,0x2F,0xF5,0x4F,0xE6,0xF0,0x47,0x79,0x0F,0x0F,0x0F,0x0F,0xE6,0x0F,0xB0,
   0x32,0xFF,0xFF,0xDB,0xA8,0xF5,0xE6,0xF0,0x47,0x0F,0x0F,0x0F,0x0F,0xB0,0xD3,0xA8,
   0x2A,0x03,0x90,0xED,0x5B,0x05,0x90,0xEB,0xED,0x52,0x44,0x4D,0x21,0x80,0x90,0xED,
   0x5B,0x03,0x90,0xED,0xB0,0xF1,0xD3,0xA8,0xF1,0x32,0xFF,0xFF,0xFB,0xCD,0xC3,0x00,
   0xCD,0x56,0x01,0xF3,0x21,0xF0,0xFB,0x22,0xF8,0xF3,0x22,0xFA,0xF3,0x21,0x09,0x90,
   0x11,0xF0,0xFB,0x01,0x14,0x00,0xED,0xB0,0x21,0x04,0xFC,0x22,0xF8,0xF3,0xFB,0xC9,
   };

char ROM32KLoader2[108] = {
   0xC3,0x09,0x90,0x00,0x00,0x00,0x00,0x00,0x00,0xF3,0x3A,0xFF,0xFF,0x2F,0xF5,0x4F,
   0xE6,0xF0,0x47,0x79,0x0F,0x0F,0x0F,0x0F,0xE6,0x0F,0xB0,0x32,0xFF,0xFF,0xDB,0xA8,
   0xF5,0xE6,0xF0,0x47,0x0F,0x0F,0x0F,0x0F,0xB0,0xD3,0xA8,0x2A,0x03,0x90,0x7C,0xE6,
   0xF0,0xFE,0x80,0xC2,0x4B,0x90,0x21,0x4B,0x90,0x11,0x6D,0x90,0xEB,0xED,0x52,0x44,
   0x4D,0x21,0x4B,0x90,0x11,0x60,0xF5,0xD5,0xED,0xB0,0xC9,0x2A,0x03,0x90,0xED,0x5B,
   0x05,0x90,0xEB,0xED,0x52,0x44,0x4D,0x21,0x6C,0x90,0xED,0x5B,0x03,0x90,0xED,0xB0,
   0xF1,0xD3,0xA8,0xF1,0x32,0xFF,0xFF,0xF7,0x00,0x00,0x00,0x00
   };

char MultFileLoader[106] = {
   0xC3,0x1E,0x90,0x00,0x00,0x00,0x00,0x00,0x00,0x42,0x4C,0x4F,0x41,0x44,0x22,0x43,
   0x41,0x53,0x3A,0x52,0x4F,0x4D,0x33,0x32,0x4B,0x22,0x2C,0x52,0x0D,0x0D,0xF3,0x21,
   0x34,0x90,0x11,0x6B,0x90,0xEB,0xED,0x52,0x44,0x4D,0x21,0x34,0x90,0x11,0x60,0xF5,
   0xD5,0xED,0xB0,0xC9,0x21,0xF0,0xFB,0x22,0xF8,0xF3,0x22,0xFA,0xF3,0x21,0x09,0x90,
   0x11,0xF0,0xFB,0x01,0x14,0x00,0xED,0xB0,0x21,0x04,0xFC,0x22,0xF8,0xF3,0x2A,0x07,
   0x90,0xE5,0x2A,0x03,0x90,0xED,0x5B,0x05,0x90,0xEB,0xED,0x52,0x44,0x4D,0x21,0x6A,
   0x90,0xED,0x5B,0x03,0x90,0xED,0xB0,0xFB,0xC9,0x00
	};



// Main routine ----------------------------------------------------------------

int main(int argc, char *argv[])
{
   // Print title
   printf("\nCasLink Version 1.0b  Copyright (c) 1999 Alex Podrezov. All rights reserved.\n");
   printf("\n This utility encodes/decodes files to be transfered through MSX cassette");
   printf("\n interface by any sound I/O capable hardware - PC sound card or CD player.\n");
   printf("\n Supported formats : BAS (tokenized/ASCII), BIN (8/16/24/32k), ROM (8/16/32k),");
   printf("\n                     RAW data files.\n");


	//Output help hint
   if (argc == 1)
   {
   	printf("\n To get help on command line options use /? key...\n\n");
      return 0;
      }

   // Check command line parameters
   if (*(unsigned char*)(argv[1]) != '/') ExitOnError(1);
   if (*(unsigned char*)(argv[1] + 1) == '?')
   {
   	PrintHelp();
      return 0;
      }
   else if (*(unsigned char*)(argv[1] + 1) == 'D' || *(unsigned char*)(argv[1] + 1) == 'd') EncodeDecode = 3;	// Decode from WAV
   else if (*(unsigned char*)(argv[1] + 1) == 'E' || *(unsigned char*)(argv[1] + 1) == 'e' || *(unsigned char*)(argv[1] + 1) == 'M' || *(unsigned char*)(argv[1] + 1) == 'm')
   {
      // Check if filename is missing
		EncodeDecode = 1;	// Encode to WAV ?
		if (argc < 2) ExitOnError(2);
		if (strlen(argv[2]) > 12 || *(unsigned char*)(argv[2] + 1) < '0' || *(unsigned char*)(argv[1] + 1) > 'z') ExitOnError(1);

      // Uppercase the string and copy input/output file name to buffers
      for (i = 0; i < strlen(argv[2]); i++)
      {
      	*(unsigned char*)(argv[2]+i) = (unsigned char)toupper(*(unsigned char*)(argv[2]+i));
			(unsigned char)InputFileName[i] = *(unsigned char*)(argv[2]+i);
         }
      (unsigned char)InputFileName[i] = '\0';
		for (i = 0; i < 8; i++)
      {
      	if ((unsigned char)InputFileName[i] == '.') break;
			(unsigned char)OutputFileName[i] = (unsigned char)InputFileName[i];
         }
      (unsigned char)OutputFileName[i] = '\0';
      if (EncodeDecode < 3) strcat(OutputFileName, ".WAV");	  // Add extension

	   if (*(unsigned char*)(argv[1] + 1) == 'M' || *(unsigned char*)(argv[1] + 1) == 'm')
	   {
	      // Check if more filenames are missing
	   	EncodeDecode = 2;	// Encode to WAV
			if (argc < 3) ExitOnError(2);
			k = argc;
			if (k > 12) k = 12;
			for (l = 3; l < k; l++)	if (strlen(argv[l]) > 12 || *(unsigned char*)(argv[l] + 1) < '0' || *(unsigned char*)(argv[l] + 1) > 'z') ExitOnError(1);
			for (l = 0; l < argc; l++) FileNamePointer[l] = argv[l];
	   	}
      }

   // In other cases exit
   else ExitOnError(1);

   // Ask signal level and MSX speed interface
	if (EncodeDecode < 3)
   {
		printf("\n  Specify MSX cassette interface speed: [1 - 1200bps, 2 - 2400bps] - ");
      ch = getch();
      if (ch == '2') MSXInterfaceSpeed = 2400;
      else if (ch == '1') MSXInterfaceSpeed = 1200;
      else ExitOnError(3);
      printf("%d", MSXInterfaceSpeed);
		printf("\n  Specify WAV signal level: [10 (lowest) - 99 (highest)] - ");
      ch = getch();
		SignalLevel = atoi(&(char)ch);
      ch = getch();
      SignalLevel = SignalLevel * 10 + atoi(&(char)ch);
      if (!SignalLevel || SignalLevel < 10) ExitOnError(3);
		printf("%d\n", SignalLevel);
      if (SignalLevel == 99) SignalLevel = 100;

	   // Make primary preparations
		PrimarySetup();
   	}

   // Encode or decode file(s)
   if (EncodeDecode == 1) EncodeSingleFile();
   else if (EncodeDecode == 2) EncodeMultipleFiles(argc);
   else DecodeFile();

   // Cleanup
   CleanUp();

   // Final message
   printf("\n Operation completed successfully! Please send your comments to CasLink author:");
   printf("\n <podrezov@hotmail.com> or to ICQ UIN 1253313.\n\n");

	return 0;
	}


// Encode and write data byte to WAV file --------------------------------------

int WriteDataByte(unsigned char ch)
{
	int i, j;
   char BitMask = 1;

  	BitStream[0] = (unsigned char)STARTBIT;
  	BitStream[9] = (unsigned char)STOPBIT;	// Fill start and stopbits
  	BitStream[10] = (unsigned char)STOPBIT;
   for (i = 1; i < 9; i++)
   {
   	if (ch & BitMask) BitStream[i] = 1;	// Fill bit stream
      else BitStream[i] = 0;
      BitMask = BitMask << 1;
      }

   for (i = 0; i < sizeof(BitStream); i++)
   {
		if ((unsigned char)BitStream[i] == 0) for (j = 0; j < BitEncodingLength; j++) (unsigned char)OutputMemPointer[i * BitEncodingLength + j] = ZeroBit[SpeedScale][j];
      else for (j = 0; j < BitEncodingLength; j++) (unsigned char)OutputMemPointer[i * BitEncodingLength + j] = SetBit[SpeedScale][j];
      }

   // Save one encoded byte to file
	OutputFileLength = (size_t)(BitEncodingLength * sizeof(BitStream));
	BytesWritten = fwrite(OutputMemPointer, sizeof(unsigned char), OutputFileLength, OutputFileHandle);
	if (BytesWritten != OutputFileLength) return 1;

	return 0;
	}


// Abnormal termination routine ------------------------------------------------

void ExitOnError(int ErrorNum)
{
	// Clean up
   CleanUp();

   // Print error message and exit
   if (!ErrorNum) exit(255);
   printf("\n\n");
   printf(ErrorList[ErrorNum-1]);
   exit(ErrorNum);
   }


// Clean up routine ------------------------------------------------------------

void CleanUp(void)
{
	if (InputMemPointer) free(InputMemPointer);
	if (OutputMemPointer) free(OutputMemPointer);
   if (InputFileHandle) fclose(InputFileHandle);
   if (OutputFileHandle) fclose(OutputFileHandle);
   }


// Encode input file to WAV sound file routine ---------------------------------

void EncodeSingleFile(void)
{
   // Open, check size, allocate memory and read input file
   InputFileHandle = fopen(InputFileName, "rb");
   if (!InputFileHandle) ExitOnError(6);
	if (fseek(InputFileHandle, 0, SEEK_END)) ExitOnError(7);
   InputFileLength = (size_t)ftell(InputFileHandle);
   if (fseek(InputFileHandle, 0, SEEK_SET)) ExitOnError(7);
	if (InputFileLength < (long)MINENCINPUTFILELENGTH || InputFileLength > (long)MAXENCINPUTFILELENGTH) ExitOnError(8);
   InputMemPointer = (unsigned char*) malloc(InputFileLength);
   if (!InputMemPointer) ExitOnError(5);
   BytesRead = fread(InputMemPointer, sizeof(unsigned char), InputFileLength, InputFileHandle);
   if (BytesRead != InputFileLength) ExitOnError(9);

   // Check program type and make necessary adjustments
   if ((unsigned char)InputMemPointer[0] == 0xff)
   {
   	FileType = 1;			// tokenized basic program
      ExtraBytes = 7;
      FileOffset = 1;
      MoreExtraBytes = 0L;
      }
   else if ((unsigned char)InputMemPointer[0] == 0xfe)
   {
   	FileType = 2;			// BIN program
      ExtraBytes = 0;
      FileOffset = 1;
      MoreExtraBytes = 0L;
      }
   else if ((unsigned char)InputMemPointer[0] == 'A' && (unsigned char)InputMemPointer[1] == 'B')
   {
   	FileType = 3;			// ROM image
      FileOffset = 0;
		if (InputFileLength < 25000L)
      {
			if ((unsigned char)InputMemPointer[3] < 0x80) ExtraBytes = sizeof(ROMLoader) + 6;
			else ExtraBytes = 6;
	      MoreExtraBytes = 0L;
         }
      else
      {
			MoreExtraBytes = (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength) + (DWORD)((DWORD)WavSampleRate * (DWORD)MiddlePauseLength) +
         	(DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength + (DWORD)WavSampleRate * (DWORD)ShortHeaderLength) +
					(DWORD)(((DWORD)sizeof(BinFileHeaderID) + (DWORD)CASFILENAMELENGTH - (DWORD)FileOffset) * (DWORD)SampleScale * (DWORD)sizeof(BitStream) * (DWORD)BitEncodingLength);
			ExtraBytes = sizeof(ROM32KLoader1) + sizeof(ROM32KLoader2) + 12;
      	}
		}
   else
   {
		for (i = 0; i < InputFileLength; i++)
      {
      	if ((unsigned char)InputMemPointer[i] >= ' ' || (unsigned char)InputMemPointer[i] <= 2 || (unsigned char)InputMemPointer[i] == 0xa || (unsigned char)InputMemPointer[i] == 0xd || (unsigned char)InputMemPointer[i] == 0x1a) continue;
         else ExitOnError(10);
         }
   	FileType = 4;			// ASCII basic file or data file
      ExtraBytes = 256;
      FileOffset = 0;
      MoreExtraBytes = 0L;
   	}

   // Check if a valid ROM
   if (FileType == 3)
   {
		ch = (unsigned char)InputMemPointer[3];
	   ch = ch & 0xf0;
   	if (InputFileLength > 25000L && ch >= 0xD0) ExitOnError(14);
      }


   // Print file format
   printf("\n Input File Format - ");
   printf(FileFormat[FileType - 1]);
   printf("\n");

   // Fill WAV header
	strcpy(WavHeader.WavHeaderID, "RIFF");
	strcpy(WavHeader.MMFileType, "WAVE");
	strcpy(WavHeader.FormatID, "fmt ");
  	strcpy(WavHeader.DataID, "data");
   WavHeader.Dummy = (WORD) 1;
   WavHeader.FormatTag = (WORD) 1;	// PCM
   WavHeader.NumChannels = (WORD) 1; // mono always
   WavHeader.BitsPerSample = (WORD) 8;	// 8 bit always
   WavHeader.SamplesPerSec = (DWORD)WavSampleRate;
   WavHeader.BytesPerSec = (DWORD)((DWORD)WavSampleRate * (DWORD)(WavHeader.BitsPerSample / 8) * (DWORD)WavHeader.NumChannels);
	WavHeader.FormatSize = (DWORD)(sizeof(WavHeader.FormatTag) + sizeof(WavHeader.NumChannels) + sizeof(WavHeader.BitsPerSample) +
   	sizeof(WavHeader.SamplesPerSec) + sizeof(WavHeader.BytesPerSec) + sizeof(WavHeader.Dummy));
	if (FileType == 1) FileHeaderIDLength = sizeof(BasicFileHeaderID);
   else if (FileType == 2 || 	FileType == 3) FileHeaderIDLength = sizeof(BinFileHeaderID);
	if (FileType == 4) FileHeaderIDLength = sizeof(DataFileHeaderID);
	if (FileType != 4)
   {
	   WavHeader.PureSampleLength = (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength + (DWORD)WavSampleRate * (DWORD)MiddlePauseLength + (DWORD)WavSampleRate * (DWORD)FinalPauseLength) +	// Length of pauses
   		(DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength + (DWORD)WavSampleRate * (DWORD)ShortHeaderLength) +	// Length of long and short headers
   			(DWORD)(((DWORD)FileHeaderIDLength + (DWORD)CASFILENAMELENGTH + (DWORD)InputFileLength + (DWORD)ExtraBytes - (DWORD)FileOffset) * (DWORD)SampleScale * (DWORD)sizeof(BitStream) * (DWORD)BitEncodingLength) +	// Length of data areas
	            MoreExtraBytes;
      }
   else
   {
		b = 1;
		if (InputFileLength > 256)
      {
			InputFileLengthTmp = InputFileLength;
			for (b = 2; ; b++)
      	{
      		InputFileLengthTmp = InputFileLengthTmp - 256;
	         if (InputFileLengthTmp <= 256) break;
   	      }
         }
	   WavHeader.PureSampleLength = (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength + (DWORD)WavSampleRate * (DWORD)MiddlePauseLength + (DWORD)WavSampleRate * (DWORD)FinalPauseLength) +	// Length of pauses
   		(DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength + (DWORD)WavSampleRate * (DWORD)ShortHeaderLength * (DWORD)b) +	// Length of long and all short headers
   			(DWORD)(((DWORD)FileHeaderIDLength + (DWORD)CASFILENAMELENGTH + (DWORD)InputFileLength + (DWORD)ExtraBytes - (DWORD)FileOffset) * (DWORD)SampleScale * (DWORD)sizeof(BitStream) * (DWORD)BitEncodingLength) +	// Length of data areas
	            MoreExtraBytes;
   	}
   WavHeader.SampleLength = (DWORD)((DWORD)WavHeader.PureSampleLength + (DWORD)WAVHEADEROFFSETVALUE);

   // Check if file already exists
   OutputFileHandle = fopen(OutputFileName, "rb");
	if (OutputFileHandle)
   {
	   printf("\n WARNING: File with this name already exists. Overwrite ? [Y/N] - ");
      ch=getch();
      if (toupper(ch) == 'N') ExitOnError(11);
      printf("%c\n", toupper(ch));
	   fclose(OutputFileHandle);
      }
   printf("\n Creating output file... Please wait...\n");

   // Write output file header (WAV)
   OutputFileHandle = fopen(OutputFileName, "wb");
   if (!OutputFileHandle) ExitOnError(12);
   BytesWritten = fwrite(&WavHeader, sizeof(unsigned char), sizeof(WavHeader), OutputFileHandle);
   if (BytesWritten != sizeof(WavHeader)) ExitOnError(13);


   // Encode first pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }

   // Create and Encode long header
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength / (DWORD)BitEncodingLength); j++)
   {
	   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
   	if (BytesWritten != BitEncodingLength) ExitOnError(13);
      }

	// Prepare file name for tape header
   for (i = 0; i < CASFILENAMELENGTH; i++)
   {
   	if ((unsigned char)InputFileName[i] != '.' && (unsigned char)InputFileName[i]) (unsigned char)NameBuffer[i] = (unsigned char)InputFileName[i];
      else break;
      }
   for ( ; i < CASFILENAMELENGTH; i++) (unsigned char)NameBuffer[i] = ' ';

   // Encode file header data
	if (FileType == 1)
   {
   	for (i = 0; i < sizeof(BasicFileHeaderID); i++) if (WriteDataByte(BasicFileHeaderID[i])) ExitOnError(13);	// Basic ID
      }
   else if (FileType == 2 || FileType == 3)
   {
   	for (i = 0; i < sizeof(BinFileHeaderID); i++) if (WriteDataByte(BinFileHeaderID[i])) ExitOnError(13);	// BIN/ROM ID
      }
   else if (FileType == 4)
   {
   	for (i = 0; i < sizeof(DataFileHeaderID); i++) if (WriteDataByte(DataFileHeaderID[i])) ExitOnError(13);	// Data ID
      }

	// Encode file name
	for (i = 0; i < CASFILENAMELENGTH; i++)
   {
   	if (WriteDataByte(NameBuffer[i])) ExitOnError(13);	// File name
		(unsigned char)ROM32KLoader1[i + 0x13] = (unsigned char)NameBuffer[i];
      }

   // Encode second pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)MiddlePauseLength); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }

   // Encode BAS and BIN files
   if (FileType < 3)
   {
		// Create and Encode short header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
		{
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
	  		if (BytesWritten != BitEncodingLength) ExitOnError(13);
		   }

	   // Encode whole program data
		for (i = FileOffset; i < InputFileLength; i++)
	   {
	   	ch = (unsigned char)InputMemPointer[i];	// Retrieve data byte
	      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	      }
      }

   // Encode ROM files smaller than 30k
   else if (FileType == 3 && InputFileLength < 25000L)	// ROM file
   {

		// Create and Encode short header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
		{
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
	  		if (BytesWritten != BitEncodingLength) ExitOnError(13);
		   }

		ch = (unsigned char)InputMemPointer[3];
		if ((unsigned char)ch < 0x80)
      {
      	// Set ROM loader addresses
			ch = ch & 0xf0;
			a = (size_t)ch;
         a = a << 8;
	     	ROMLoader[3] = 0;
        	ROMLoader[4] = (char)(a >> 8);
         b = a + InputFileLength;
	     	ROMLoader[5] = (char)b;
        	ROMLoader[6] = (char)(b >> 8);
      	ROMLoader[7] = (char)InputMemPointer[2];
	      ROMLoader[8] = (char)InputMemPointer[3];

			// Set binary start addresses
			AddressBuffer[0] = 0;
			AddressBuffer[1] = 0x90;
         a = (size_t)((size_t)sizeof(ROMLoader) + (size_t)InputFileLength + (size_t)0x9000 - 1);
			AddressBuffer[2] = (char)a;
			AddressBuffer[3] = (char)(a >> 8);
			AddressBuffer[4] = 0;
			AddressBuffer[5] = 0x90;

		   // Encode 6 bytes of addresses
			for (i = 0; i < sizeof(AddressBuffer); i++)
	   	{
		   	ch = (unsigned char)AddressBuffer[i];		// Retrieve data byte
		      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	   	   }

		   // Encode binary ROM loader
			for (i = 0; i < sizeof(ROMLoader); i++)
	   	{
		   	ch = (unsigned char)ROMLoader[i];		// Retrieve data byte
		      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	   	   }
			}

      else
      {
          // Encode 6 bytes of addresses
			ch = ch & 0xf0;
			a = (size_t)ch;
         a = a << 8;
			AddressBuffer[0] = 0;
			AddressBuffer[1] = (char)(a >> 8);
         b = a + InputFileLength - 1;
			AddressBuffer[2] = (char)b;
			AddressBuffer[3] = (char)(b >> 8);
			AddressBuffer[4] = (char)InputMemPointer[2];
			AddressBuffer[5] = (char)InputMemPointer[3];

		   // Encode 6 bytes of addresses
			for (i = 0; i < sizeof(AddressBuffer); i++)
	   	{
		   	ch = (unsigned char)AddressBuffer[i];		// Retrieve data byte
		      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	   	   }
      	}

	   // Encode whole ROM data
		for (i = FileOffset; i < InputFileLength; i++)
	   {
	   	ch = (unsigned char)InputMemPointer[i];	// Retrieve data byte
	      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	      }
   	}

   // Encode ROM files larger than 33k
   else if (FileType == 3 && InputFileLength > 25000L)	// ROM file
   {

		// Create and Encode short header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
		{
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
	  		if (BytesWritten != BitEncodingLength) ExitOnError(13);
		   }

     	// Set ROM loader addresses
		ch = (unsigned char)InputMemPointer[3];
		if (ch > 0x80) ch = (unsigned char)0x40;
      ch = ch & 0xf0;
      a = (size_t)ch;
      a = a << 8;
      ROM32KLoader1[3] = 0;
      ROM32KLoader1[4] = (char)(a >> 8);
      b = a + (size_t)16384;
      ROM32KLoader1[5] = (char)b;
      ROM32KLoader1[6] = (char)(b >> 8);
      ROM32KLoader1[7] = (char)InputMemPointer[2];
      ROM32KLoader1[8] = (char)InputMemPointer[3];

      // Set binary start addresses
      AddressBuffer[0] = 0;
      AddressBuffer[1] = 0x90;
      a = (size_t)((size_t)sizeof(ROM32KLoader1) + (size_t)16384 + (size_t)0x9000 - 1);
      AddressBuffer[2] = (char)a;
      AddressBuffer[3] = (char)(a >> 8);
      AddressBuffer[4] = 0;
      AddressBuffer[5] = 0x90;

      // Encode 6 bytes of addresses
      for (i = 0; i < sizeof(AddressBuffer); i++)
      {
      	ch = (unsigned char)AddressBuffer[i];		// Retrieve data byte
         if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
         }

      // Encode first binary 32k ROM loader
      for (i = 0; i < sizeof(ROM32KLoader1); i++)
      {
      	ch = (unsigned char)ROM32KLoader1[i];		// Retrieve data byte
         if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
			}

	   // Encode half of 32k ROM data
		for (i = FileOffset; i < (size_t)16384; i++)
	   {
	   	ch = (unsigned char)InputMemPointer[i];	// Retrieve data byte
	      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	      }

	   // Encode pause between files
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength); j++)
	   {
		   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   	if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
	      }

	   // Create and Encode long header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength / (DWORD)BitEncodingLength); j++)
	   {
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
   		if (BytesWritten != BitEncodingLength) ExitOnError(13);
	      }

      // Encode binary header and second part of 32k ROM name
   	for (i = 0; i < sizeof(BinFileHeaderID); i++) if (WriteDataByte(BinFileHeaderID[i])) ExitOnError(13);	// BIN/ROM ID
		for (i = 0; i < CASFILENAMELENGTH; i++) if (WriteDataByte(NameBuffer[i])) ExitOnError(13);	// File name

	   // Encode second pause
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)MiddlePauseLength); j++)
	   {
		   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
		   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
	      }

		// Create and encode short header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
		{
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
	  		if (BytesWritten != BitEncodingLength) ExitOnError(13);
		   }

		ch = (unsigned char)InputMemPointer[3];
		if (ch > 0x80) ch = (unsigned char)0x40;

      // Set ROM loader addresses
      ch = ch & 0xf0;
      a = (size_t)ch;
      a = a << 8;
      a = a  + (size_t)16384;
      ROM32KLoader2[3] = 0;
      ROM32KLoader2[4] = (char)(a >> 8);
      b = a + InputFileLength - (size_t)16384;
      ROM32KLoader2[5] = (char)b;
      ROM32KLoader2[6] = (char)(b >> 8);
      ROM32KLoader2[7] = (char)InputMemPointer[2];
      ROM32KLoader2[8] = (char)InputMemPointer[3];

      // Set binary start addresses
      AddressBuffer[0] = 0;
      AddressBuffer[1] = 0x90;
      a = (size_t)((size_t)sizeof(ROM32KLoader2) + (size_t)InputFileLength - (size_t)16384 + (size_t)0x9000 - 1);
      AddressBuffer[2] = (char)a;
      AddressBuffer[3] = (char)(a >> 8);
      AddressBuffer[4] = 0;
      AddressBuffer[5] = 0x90;

      // Encode 6 bytes of addresses
      for (i = 0; i < sizeof(AddressBuffer); i++)
      {
      	ch = (unsigned char)AddressBuffer[i];		// Retrieve data byte
      	if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
         }

      // Encode second binary 32k ROM loader
      for (i = 0; i < sizeof(ROM32KLoader2); i++)
      {
      	ch = (unsigned char)ROM32KLoader2[i];		// Retrieve data byte
	      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
         }

	   // Encode second half of 32k ROM data
		for (i = FileOffset + (size_t)16384; i < InputFileLength; i++)
	   {
	   	ch = (unsigned char)InputMemPointer[i];	// Retrieve data byte
	      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
	      }
		}

	// Encode data or detokenized BAS files
   else
   {
		if (InputFileLength <= 256) b = InputFileLength;
      else b = 256;

	   // Encode 256 bytes + header chunks of data
		for (i = FileOffset; ; i = i + 256)
	   {
	   	// Create and Encode short header
			for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
			{
			   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
			  	if (BytesWritten != BitEncodingLength) ExitOnError(13);
			   }

         // Encode 256 (or less) bytes
			for (a = 0; a < b; a++)
         {
		   	ch = (unsigned char)InputMemPointer[i + a];	// Retrieve data byte
		      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
		      }

         // Check if whole file is written
         if (InputFileLength <= 256) break;
			InputFileLength = InputFileLength - 256;
         if (InputFileLength <= 256) b = InputFileLength;
   		}
   	}

   // Encode finalize areas
	if (FileType == 1)
   {
   	for (i = 0; i < 7; i++) if (WriteDataByte(0)) ExitOnError(13);	// Encode 7 zeros at the end of basic program
      }
	else if (FileType == 4)
   {
   	for (i = 0; i < 256; i++) if (WriteDataByte(0x1a)) ExitOnError(13);	// Encode 256 bytes at the end of ascii program
      }


   // Encode final pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FinalPauseLength + LENGTHCORRECTION); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }

	}


// Make primary preparations

void PrimarySetup(void)
{
   // Initialize variables
  	WavSampleRate = 11025;			// 11025, 22050, 44100 Hz (8-bit, mono)
   LongHeaderLength = 20/3;
   ShortHeaderLength = 5/3;
   FirstPauseLength = 2;			// Seconds
   MiddlePauseLength = 1;				// Seconds
   FinalPauseLength = 1;			// Seconds
	SilenceByte = SILENCE;

   // Prepare scales and bit patterns according to volume
   SampleScale = WavSampleRate / 11025;   // sample scale for different sampling rates
   SpeedScale = MSXInterfaceSpeed / 1200 - 1;	// speed scale for MSX interface
	if (!SpeedScale) BitEncodingLength = 10;
   else BitEncodingLength = 5;
   for (i = 0; i < BitEncodingLength; i++)
   {
		if (ZeroBit[SpeedScale][i] == (unsigned char)LOWAMPLITUDE) ZeroBit[SpeedScale][i] = (unsigned char)(LOWAMPLITUDE + (MAXSIGNALLEVEL - SignalLevel));
      else if (ZeroBit[SpeedScale][i] == (unsigned char)HIGHAMPLITUDE) ZeroBit[SpeedScale][i] = (unsigned char)(HIGHAMPLITUDE - (MAXSIGNALLEVEL - SignalLevel));
		if (SetBit[SpeedScale][i] == (unsigned char)LOWAMPLITUDE) SetBit[SpeedScale][i] = (unsigned char)(LOWAMPLITUDE + (MAXSIGNALLEVEL - SignalLevel));
      else if (SetBit[SpeedScale][i] == (unsigned char)HIGHAMPLITUDE) SetBit[SpeedScale][i] = (unsigned char)(HIGHAMPLITUDE - (MAXSIGNALLEVEL - SignalLevel));
    	}

   // Allocate memory for output buffer
   OutputMemPointer = (unsigned char*) malloc((size_t)(BitEncodingLength * sizeof(BitStream)));
   if (!OutputMemPointer) ExitOnError(4);
   }


// Print help info

void PrintHelp(void)
{
	printf("\n");
   printf("\n Usage: CasLink <key> <filename>\n");
   printf("\n  Keys: <filename> - name of input/output file (12 characters max);");
   printf("\n        </E> - encode basic or binary file into WAV sound file;");
   printf("\n        </M> - encode multiple files (up to 10) into WAV sound file;");
   printf("\n        </D> - decode basic or binary file from WAV sound file;");
   printf("\n        </?> - this help screen.");
   printf("\n");
   printf("\n  Note: File name encoded to WAV file will be in capital letters.");
   printf("\n\n");
	}


// Encode multiple input files to WAV sound file routine -----------------------

void EncodeMultipleFiles(int k)
{
	struct {
		char FileType;
      size_t ExtraBytes;
      size_t FileOffset;
		unsigned long FileLength;
   } FileQueue[10];
   int l;

	// Collect information about all input files
   MoreExtraBytes = 1L;
	if (k > 12) k = 12;
	for (l = 2; l < k; l++)
   {
	   // Open, check size, allocate memory and read 10 bytes of input file
   	InputFileHandle = fopen(FileNamePointer[l], "rb");
	   if (!InputFileHandle) ExitOnError(6);
		if (fseek(InputFileHandle, 0, SEEK_END)) ExitOnError(7);
	   InputFileLength = (size_t)ftell(InputFileHandle);
   	if (fseek(InputFileHandle, 0, SEEK_SET)) ExitOnError(7);
		if (InputFileLength < (long)MINENCINPUTFILELENGTH || InputFileLength > (long)MAXENCINPUTFILELENGTH) ExitOnError(8);
   	InputMemPointer = (unsigned char*) malloc(10);
	   if (!InputMemPointer) ExitOnError(5);
   	BytesRead = fread(InputMemPointer, sizeof(unsigned char), 10, InputFileHandle);
	   if (BytesRead != 10) ExitOnError(9);
	   if ((unsigned char)InputMemPointer[0] == 0xff) ExitOnError(16);
		else if ((unsigned char)InputMemPointer[0] == 'A' && (unsigned char)InputMemPointer[1] == 'B') ExitOnError(15);
		else if ((unsigned char)InputMemPointer[0] == 0xfe)
      {
      	FileQueue[l-2].FileType = 2; // BIN program
			if (FileQueue[0].FileType != 1)	// no basic program is loaded first
         {
		      FileQueue[l-2].ExtraBytes = sizeof(MultFileLoader);
   		   FileQueue[l-2].FileOffset = 7;
            }
         else
         {
		      FileQueue[l-2].ExtraBytes = 0;
   		   FileQueue[l-2].FileOffset = 1;
            }
         }
		else
      {
      	FileQueue[l-2].FileType = 1; // BAS program
			if (FileQueue[0].FileType != 1) ExitOnError(17);
   	   FileQueue[l-2].ExtraBytes = 256;
      	FileQueue[l-2].FileOffset = 0;
			b = 1;
			if (InputFileLength > 256)
	      {
				InputFileLengthTmp = InputFileLength;
				for (b = 2; ; b++)
	      	{
	      		InputFileLengthTmp = InputFileLengthTmp - 256;
		         if (InputFileLengthTmp <= 256) break;
   		      }
         	}
         if (MoreExtraBytes == 1L) MoreExtraBytes = 0L;
         MoreExtraBytes = MoreExtraBytes + b;
         }

      FileQueue[l-2].FileLength = (unsigned long)InputFileLength;	 // Save length
		fclose(InputFileHandle);
      free(InputMemPointer);
      InputMemPointer = 0;
		InputFileHandle = 0;
      }

   // Fill WAV header
	strcpy(WavHeader.WavHeaderID, "RIFF");
	strcpy(WavHeader.MMFileType, "WAVE");
	strcpy(WavHeader.FormatID, "fmt ");
  	strcpy(WavHeader.DataID, "data");
   WavHeader.Dummy = (WORD) 1;
   WavHeader.FormatTag = (WORD) 1;	// PCM
   WavHeader.NumChannels = (WORD) 1; // mono always
   WavHeader.BitsPerSample = (WORD) 8;	// 8 bit always
   WavHeader.SamplesPerSec = (DWORD)WavSampleRate;
   WavHeader.BytesPerSec = (DWORD)((DWORD)WavSampleRate * (DWORD)(WavHeader.BitsPerSample / 8) * (DWORD)WavHeader.NumChannels);
	WavHeader.FormatSize = (DWORD)(sizeof(WavHeader.FormatTag) + sizeof(WavHeader.NumChannels) + sizeof(WavHeader.BitsPerSample) +
   	sizeof(WavHeader.SamplesPerSec) + sizeof(WavHeader.BytesPerSec) + sizeof(WavHeader.Dummy));
	for (l = 2; l < k; l++)
   {
		if (FileQueue[l-2].FileType == 1) FileHeaderIDLength = FileHeaderIDLength + sizeof(BasicFileHeaderID);
      else FileHeaderIDLength = FileHeaderIDLength + sizeof(BinFileHeaderID);
	   WavHeader.PureSampleLength = WavHeader.PureSampleLength + (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength + (DWORD)WavSampleRate * (DWORD)MiddlePauseLength) +	// Length of pauses
   		(DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength + (DWORD)WavSampleRate * (DWORD)ShortHeaderLength * (DWORD)MoreExtraBytes) +	// Length of long and short headers
   			(DWORD)(((DWORD)FileHeaderIDLength + (DWORD)CASFILENAMELENGTH + (DWORD)FileQueue[l-2].FileLength + (DWORD)FileQueue[l-2].ExtraBytes - (DWORD)FileQueue[l-2].FileOffset) * (DWORD)SampleScale * (DWORD)sizeof(BitStream) * (DWORD)BitEncodingLength);	// Length of data areas
      }
	WavHeader.PureSampleLength = WavHeader.PureSampleLength + (DWORD)((DWORD)WavSampleRate * (DWORD)FinalPauseLength);
   WavHeader.SampleLength = (DWORD)((DWORD)WavHeader.PureSampleLength + (DWORD)WAVHEADEROFFSETVALUE);

   // Check if file already exists
   OutputFileHandle = fopen(OutputFileName, "rb");
	if (OutputFileHandle)
   {
	   printf("\n WARNING: File with this name already exists. Overwrite ? [Y/N] - ");
      ch=getch();
      if (toupper(ch) == 'N') ExitOnError(11);
      printf("%c\n", toupper(ch));
	   fclose(OutputFileHandle);
      }
   printf("\n Creating output file... Please wait...\n");

   // Write output file header (WAV)
   OutputFileHandle = fopen(OutputFileName, "wb");
   if (!OutputFileHandle) ExitOnError(12);
   BytesWritten = fwrite(&WavHeader, sizeof(unsigned char), sizeof(WavHeader), OutputFileHandle);
   if (BytesWritten != sizeof(WavHeader)) ExitOnError(13);

   // Encode all files
	for (l = 2; l < k; l++)
   {
	   // Encode first pause
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength); j++)
	   {
		   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   	if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
	      }

	   // Create and encode long header
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength / (DWORD)BitEncodingLength); j++)
	   {
		   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
   		if (BytesWritten != BitEncodingLength) ExitOnError(13);
	      }

	   // Encode file header data
		if (FileQueue[l-2].FileType == 1)
	   {
	   	for (i = 0; i < sizeof(DataFileHeaderID); i++) if (WriteDataByte(DataFileHeaderID[i])) ExitOnError(13);	// Basic ID
	      }
	   else
	   {
	   	for (i = 0; i < sizeof(BinFileHeaderID); i++) if (WriteDataByte(BinFileHeaderID[i])) ExitOnError(13);	// BIN/ROM ID
   	   }

		// Prepare file name for tape header
   	for (i = 0; i < CASFILENAMELENGTH; i++)
	   {
   		if ((unsigned char)InputFileName[i] != '.' && (unsigned char)InputFileName[i]) (unsigned char)NameBuffer[i] = (unsigned char)InputFileName[i];
      	else break;
	      }
	   for ( ; i < CASFILENAMELENGTH; i++) (unsigned char)NameBuffer[i] = ' ';

		// Encode file name
		for (i = 0; i < CASFILENAMELENGTH; i++)
	   {
	   	if (WriteDataByte(NameBuffer[i])) ExitOnError(13);	// File name
			(unsigned char)MultFileLoader[i + 0x13] = (unsigned char)NameBuffer[i];
	      }

	   // Encode second pause
		for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)MiddlePauseLength); j++)
	   {
		   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
		   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
   	   }

	   // Open, check size, allocate memory and read input file
   	InputFileHandle = fopen(FileNamePointer[l], "rb");
	   if (!InputFileHandle) ExitOnError(6);
		if (fseek(InputFileHandle, 0, SEEK_END)) ExitOnError(7);
	   InputFileLength = (size_t)ftell(InputFileHandle);
   	if (fseek(InputFileHandle, 0, SEEK_SET)) ExitOnError(7);
		if (InputFileLength < (long)MINENCINPUTFILELENGTH || InputFileLength > (long)MAXENCINPUTFILELENGTH) ExitOnError(8);
   	InputMemPointer = (unsigned char*) malloc(InputFileLength);
	   if (!InputMemPointer) ExitOnError(5);
   	BytesRead = fread(InputMemPointer, sizeof(unsigned char), InputFileLength, InputFileHandle);
	   if (BytesRead != InputFileLength) ExitOnError(9);

      // Encode basic file
		if (FileQueue[l-2].FileType == 1)
      {
			if (FileQueue[l-2].FileLength <= (size_t)256) b = (size_t)FileQueue[l-2].FileLength;
	      else b = 256;

		   // Encode 256 bytes + header chunks of data
			for (i = FileQueue[l-2].FileOffset; ; i = i + 256)
		   {
		   	// Create and Encode short header
				for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
				{
				   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
				  	if (BytesWritten != BitEncodingLength) ExitOnError(13);
			   	}

	         // Encode 256 (or less) bytes
				for (a = 0; a < b; a++)
	         {
			   	ch = (unsigned char)InputMemPointer[i + a];	// Retrieve data byte
			      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
			      }

	         // Check if whole file is written
   	      if (FileQueue[l-2].FileLength <= 256) break;
				FileQueue[l-2].FileLength = FileQueue[l-2].FileLength - 256;
	         if (FileQueue[l-2].FileLength <= (size_t)256) b = (size_t)FileQueue[l-2].FileLength;
   			}
   		}

      // Encode binary file
      else
      {
			// Create and Encode short header
			for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
			{
		   	BytesWritten = fwrite(SetBit[SpeedScale], sizeof(unsigned char), BitEncodingLength, OutputFileHandle);
		  		if (BytesWritten != BitEncodingLength) ExitOnError(13);
			   }

			ch = (unsigned char)InputMemPointer[2];
			if ((unsigned char)ch < 0x80) ExitOnError(15);

      	// Copy BIN start addresses to loader
         for (j = 0; j < 6; j++) MultFileLoader[j + 3] = InputMemPointer[j + 1];
         if (l == (k - 1)) for (j = 0; j < 26; j++) (char)MultFileLoader[j + 0x34] = '\0';	// patch loader if last file is output

			if (FileQueue[0].FileType != 1)	// no basic program is loaded first
         {
				// Set binary start addresses
				AddressBuffer[0] = 0;
				AddressBuffer[1] = 0x90;
	         a = (size_t)((size_t)sizeof(MultFileLoader) + (size_t)FileQueue[l-2].FileLength + (size_t)0x9000 - FileQueue[l-2].FileOffset - 1);
				AddressBuffer[2] = (char)a;
				AddressBuffer[3] = (char)(a >> 8);
				AddressBuffer[4] = 0;
				AddressBuffer[5] = 0x90;

			   // Encode 6 bytes of addresses
				for (i = 0; i < sizeof(AddressBuffer); i++)
	   		{
		   		ch = (unsigned char)AddressBuffer[i];		// Retrieve data byte
		      	if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
		   	   }

			   // Encode binary ROM loader
				for (i = 0; i < sizeof(MultFileLoader); i++)
	   		{
		   		ch = (unsigned char)MultFileLoader[i];		// Retrieve data byte
		      	if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
		   	   }
            }

		   // Encode whole BIN file from specified offset
			for (i = FileQueue[l-2].FileOffset; i < FileQueue[l-2].FileLength; i++)
	   	{
	   		ch = (unsigned char)InputMemPointer[i];	// Retrieve data byte
		      if (WriteDataByte(ch)) ExitOnError(13);	// Encode data byte
		      }
      	}

	   // Encode finalize areas
		if (FileQueue[l-2].FileType == 1)
	   {
	   	for (i = 0; i < 256; i++) if (WriteDataByte(0x1a)) ExitOnError(13);	// Encode 256 bytes at the end of ascii program
	      }

      // Cleanup before the next file is processed
		fclose(InputFileHandle);
      free(InputMemPointer);
      InputMemPointer = 0;
		InputFileHandle = 0;
   	}

   // Encode final pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FinalPauseLength + LENGTHCORRECTION); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(unsigned char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }
	}



// Decode file from WAV sound file routine -------------------------------------

void DecodeFile(void)
{

	}

